
/*
 *   THIS FILE IS UNDER RCS - DO NOT MODIFY UNLESS YOU HAVE
 *   CHECKED IT OUT USING THE COMMAND CHECKOUT.
 *
 *    $Id: edata2ew.c 2021-06-01 19:47:08Z mbertoni $
 *
 *    Revision history:
 *     $Log$
 *
 * 
 *     Revision 1.2a  2021/06/08 11:48:19  mbertoni
 *     	- added flag gcfg_WNRO for debug on timestamp issues
 *
 *     Revision 1.1a  2021/06/02 13:48:19  mbertoni
 *     	- added connection to Earthworm (connection to shared memory)
 *		- added flag for debug levels
 *		- rewritten parsing functions for channel and location names
 *		- added control for channel/locatin names read from configuration file and channels number read from DAT packet  

 *     Revision 1.0  2021/06/01 13:48:19  mbertoni
 *     Initial revision, for debug purpuse only: just reading data from serial port and simulating transmission to Earthworm shared memory.
 *
 *
 *
 */
 
 

/*  edata2ew.c:  Main (top-level) module for Earthdata-to-Earthworm   */
/*                                                                    */
/*     edata2ew = Earthdata-to-Earthworm Interface Module,            */
/*                                                                    */



/*
 * edata2ew.c:  Basic earthworm module to read data from Earthdata digitizer through serial.
 *              Code taken from template.c and serial routines from k2ew module source files. 
 *             	This module mainly opens serial port, reads data from serial port, then check starts for Earthdata serial protocol. 
 *				Once checksum pass MOD data parameters are read (digitizer type, version, serial number, number of channels, sampling rate...).
 *				After data extraction from DAT into local buffer (_s_Stream_buffer) single streams are filled, then sent to shared memory.
 */

//#define MODULE_RELVER_STR "1.0a"			// release iniziale
//#define MODULE_RELVER_STR "1.0b"        	// using date23 instead date20
//#define MODULE_RELVER_STR "1.1a"        	// starting using Earthworm
//#define MODULE_RELVER_STR "1.2a"        	// added flag for WNRO offset
//#define MODULE_RELVER_STR "1.2b"        	// increased debug level to 5 for debug purposes
//#define MODULE_RELVER_STR "1.3a"        	// let's try to read timestamp from $GPRMC message if ltime is 0

#define MODULE_RELVER_STR "1.3b"        	// emulazione $GPRMC (for debug purposes only)



/*

	Still to do:
	- get GPS position 
	- get PLL values
	- get temp
	- get aux channel/s

*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <earthworm.h>
#include <kom.h>
#include <transport.h>
#include <lockfile.h>
#include <termios.h> 
#include <fcntl.h> 
#include <sys/ioctl.h>
#include <sys/stat.h> 
#include <errno.h>
#include <transport.h>
#include <trace_buf.h>
#include <time_ew.h>	// gmtime_ew


#include <chron3.h>	// per date20
#include <float.h>	// per DBL_MAX

#include "edata.h"
#include "globalvars.h"









#define standalone	// debug mode, not working with Earthworm (for debug purposes), just reading data from serial line
					// if you need to send data to Earthworm ring, then comment "standalone"



// DO NOT ENABLE EMULATED_GPRMC!!!!!!!!!
//#define EMULATED_GPRMC	// for debug purposes: read ltime (epoch timestamp), creating $GMRMC fake string, clearing ltime fiels, taking time from $GPRMC



/* Functions in this source file 
 *******************************/
int  edata2ew_ReadConfigFile(char *);	// read configuration file
void edata2ew_PrintConfig(void);		// print current configuration


char edata_tty_init(char *sPort, unsigned int bitrate);
static int edata_GetMoreBytes(int to_read, int toutms);



void  edata2ew_lookup  ( void );
void  edata2ew_status  ( unsigned char, short, char * );
void  edata2ew_log_cfgparams(void);









/*int edata_init_io(s_tty_io *gen_io, int toutms);
int edata_close_io(void);
static int GetMoreBytes(int to_read, int toutms);
int k2c_recv_buff(unsigned char *rbuff, int datalen, int toutms);
int k2c_rcvflg_tout(int toutms);
int k2c_flush_recv(int toutms);
*/


static  SHM_INFO  Region;      /* shared memory region to use for i/o    */

#define   MAXLOGO   2
MSG_LOGO  GetLogo[MAXLOGO];    /* array for requesting module,type,instid */
short     nLogo;
pid_t     MyPid;               /* for restarts by startstop               */

MSG_LOGO  SendMsg;

//#define BUF_SIZE 60000          /* define maximum size for an event msg   */
//static char Buffer[BUF_SIZE];   /* character string to hold event message */
  

#define BUF_SIZE 4096          /* define maximum size for an event msg   */
static char msgbuf[BUF_SIZE];   /* character string to hold event message */


int ttyfd = -1; // file descriptor porta seriale 
// strutture
struct termios tty_options;

#define RB_LEN 256 
static char recv_buff[RB_LEN];

static unsigned char data_buff[512];


char RxBuffer[MaxRxMsgLength];


char *RxBuffer_ptr;
char *RxBuffer_ptr_limit;
unsigned char *data_buff_ptr;
unsigned char *SUM_ptr;
unsigned char *DATA_ptr;
unsigned short DataLen_B;

char *twrite_ptr;
char *tread_ptr;
char *printMOD_ptr;
unsigned short printMOD_n;

time_t epoch;


u_Sample_i4 preo;	// convert int32 <--> char
u_sysStatus	SystemStatus;

s_MOD	block_MOD;
s_MDE	block_MDE;
s_DATA	block_DATA;
s_SUM	block_SUM;
s_Config DeviceConfig;

s_TraceBlock DummyTrace;		// dummy trace to send data to Earthworm ring (è una struttura TRACE2_HEADER+databuffer[4096])
TracePacket   wf;               // trace packet 


static unsigned char nSCNL;	// caontatore canali attivi
static unsigned char nLOC;	// contatore location attive

static int rb_read = 0;  /* Where to start reading from recv_buff */
static int rb_count = 0; /* Number of bytes in recv_buff; note: we only write *
                       * (by calling recv()) into recv_buff when it is empty */
int edata_recv_buff(unsigned char *rbuff, int datalen, int toutms);
//static int GetMoreBytes(int to_read, int toutms);
static int edata_GetMoreBytes(int to_read, int toutms);
void edata2ew_SendStatus( unsigned char type, short ierr, char *note );
  
  
  
  

char GPRMC_parsing ( char *ptr);  
  
/* Things to read or derive from configuration file
 **************************************************/
static char    RingName[MAX_RING_STR];        /* name of transport ring for i/o    */
static char    MyModName[MAX_MOD_STR];       /* speak as this module name/id      */
static int     LogSwitch;           /* 0 if no logfile should be written */
static long    HeartBeatInterval;   /* seconds between heartbeats        */

/* Things to look up in the earthworm.h tables with getutil.c functions
 **********************************************************************/
static long          RingKey;       /* key of transport ring for i/o     */
static unsigned char InstId;        /* local installation id             */
static unsigned char MyModId;       /* Module Id for this program        */
static unsigned char TypeHeartBeat; 
static unsigned char TypeError;
static unsigned char TypeTrace2; 

/* Error messages used by template 
 *********************************/
#define  ERR_MISSMSG       0   /* message missed in transport ring       */
#define  ERR_TOOBIG        1   /* retreived msg too large for buffer     */
#define  ERR_NOTRACK       2   /* msg retreived; tracking limit exceeded */
static char  Text[150];        /* string for log/error messages          */


// global configuration variables
//s_gen_io gen_io;
//s_config station_config;
				// station configuration
int gcfg_commtimeout_itvl = 5000;      // communication timeout (millisecs) 
static char msg_txt[180];
int gcfg_debug = 0;                    // edata2ew debug level 

char gcfg_WNRO = 0;                    // edata2ew offset for WNRO


void printHexData (unsigned char *ptr, unsigned short n);	// for debug

// A buffer for recv functions and its counters 
//static unsigned char recv_buff[RB_LEN];
#define K2C_MAX_FLUSH 1000       /* max count value for 'k2c_flush_recv' */









s_SysCfg			SystemConfig;
s_Stream_buffer		Traces;			// temp buffer for incoming data

unsigned int u;	// contatore caratteri letti
unsigned int rx_count;
double dummy_double;
double dT;

static char   Str1970[18] = "19700101000000.00";
static char    strDate[26];       // per la stampa della data
double Sec1970;				
unsigned char 	NofC, NofB; 
unsigned int Fs;
int gcfg_debug;



char RMC_string[128];
void GPRMC_emulation ( char *ptr);




main( int argc, char **argv )
{
	time_t		timeNow;          // current system time                 
	time_t      timeLastBeat;     // time last heartbeat was sent  
	long      	recsize;          // size of retrieved message     
	MSG_LOGO  	reclogo;          // logo of retrieved message     
	int       	res;
 
	time_t		time_ltime;          // epoch time got from digitizer
	struct tm	*tm_NMEA;				// struct tm for NMES
	char 		MyNMEA[256];
	unsigned char locChk;
  
  
  
	static int rc,cb_rc, rb, rb_counter;

	epoch = 0;




/*
	sprintf(RMC_string, "$GPRMC,161229.487,A,3723.2475,N,12158.3416,W,0.13,309.62,120598, ,*30");
	sprintf(RMC_string, "$GPRMC,184946,A,5227.8703,N,00236.2167,W,000.0,000.0,251001,004.4,W*7C");

	GPRMC_parsing (RMC_string);
	exit(0);
*/




	
	SystemStatus.Val = 0;	// reset all system status flag
	Sec1970 = julsec17( Str1970 );	// Sec1970 è un fattore di conversione che utilizzo per le funzioni di tempo in Ew
	
	NofC = NofB = 0;
	Fs = 0;
	gcfg_debug = 0;
	nSCNL = nLOC = 0;	// caontatore canali/location attivi

	// Check command line arguments 
	if ( argc != 2 )
	{
        fprintf( stderr, "Usage: edata2ew <configfile> (e.g. edata2ew edata2ew_tty.d)\n" );
        exit( 0 );
	}
   

  
   
	// Initialize name of log-file & open it 
	logit_init( argv[1], 0, 256, 1 );
   
   
	logit("","edata2ew ver.%s\r\n", MODULE_RELVER_STR);
	
	// Read the configuration file(s)
	if (edata2ew_ReadConfigFile(argv[1]) == -1)  // process configuration file 
	{
		fprintf(stderr, "edata2ew: error processing config file <%s>\n", argv[1]);
		return 1;    // if error then exit program (error reported in function) 
	}
   
	logit( "" , "%s: Read command file <%s>\n", argv[0], argv[1] );
	if (gcfg_debug > 0)
	{
		printf("[edata2ew] Read configuration file...\r\n");
	}
	
	if (gcfg_debug > 0)
	{
		logit("","debug level: %d\r\n", gcfg_debug);
#ifdef standalone		
		logit("","standalone mode: ACTIVE\r\n");

#endif
	}
	else
	{
		logit("","debug level: disabled\r\n");
	}
	
	
	
	
#ifndef	standalone
	// Look up important info from earthworm*.d tables
	edata2ew_lookup();
	
	// format MSG_LOGO for output trace (in this example MSG_LOGO will not change
	SendMsg.instid = InstId;	// write my IntID
	SendMsg.mod = MyModId;		// write module ID of mycustom module
	SendMsg.type = TypeTrace2;	// I send data in TraceBuf2 format
#endif


	// Reinitialize logit to desired logging level 
	logit_init( argv[1], 0, 256, LogSwitch );

  
	// Get process ID for heartbeat messages 
	MyPid = getpid();
	if( MyPid == -1 )
	{
		logit("e","edata2ew: Cannot get pid. Exiting.\n");
		exit (-1);
	}



	// devo andare a inizializzare qualcos'altro?
	edata2ew_PrintConfig();




#ifndef	standalone
	// Attach to Input/Output shared memory ring 
	tport_attach( &Region, RingKey );
	logit( "", "edata2ew: Attached to public memory region %s: %ld\n", RingName, RingKey );	
	
	// Flush the transport ring
	while( tport_getmsg( &Region, GetLogo, nLogo, &reclogo, &recsize, msgbuf, sizeof(msgbuf)-1 ) != GET_NONE );	// versione template.c, qui uso msgbuf da 4096 bytes
	
	
	// Force a heartbeat to be issued in first pass thru main loop
	timeLastBeat = time(&timeNow) - HeartBeatInterval - 1;
	
#endif

	// formatting header with known values...
	memcpy((char *)&DummyTrace.header.sta, (char *)&SystemConfig.StationCode, 5);	// station
	// channel, da finire
	memcpy((char *)&DummyTrace.header.net, (char *)&SystemConfig.NetworkName, 3);	// network (2 + NULL)	
	// fake location, da finire
	sprintf(DummyTrace.header.loc, "--");	// fake location
	sprintf(DummyTrace.header.datatype, "i4");					// 32bit signed integer
        DummyTrace.header.version[0] = TRACE2_VERSION0;
        DummyTrace.header.version[1] = TRACE2_VERSION1;



	


	// going to open serial port....


	if ((res = edata_tty_init(SystemConfig.tty.name, SystemConfig.tty.bitrate)) != CODE_NO_ERROR)
	{
		printf("\tERROR opening serial port <%s>\r\n", SystemConfig.tty.name);
		exit (-1);
	}
	
	
	
	unsigned char t;
	
	
	
#ifndef	standalone
	while( (tport_getflag( &Region ) != TERMINATE)  && (tport_getflag( &Region ) != MyPid))// && (!SystemStatus.flag.SystemBlocked))	
#else
	while(1)
#endif
	{
	
		
#ifndef	standalone	
		// check if it's time to send heartbeat
		if( HeartBeatInterval  &&  (time(&timeNow)-timeLastBeat >= HeartBeatInterval)) // && (!SystemStatus.flag.SystemBlocked) )
		{
			timeLastBeat = time(&timeNow);
			edata2ew_SendStatus( TypeHeartBeat, 0, "" );
			//printf("sending hb keepalive message...\r\n");	
		}
#endif	
	
	
	
	
	
		
		rx_count = 0;
		u = 0;	// inizializzo contatore caratteri letti
		RxBuffer_ptr = RxBuffer;
		data_buff_ptr = data_buff;
		while((res = edata_recv_buff(data_buff, 400, 200)) > 0)
		{
			
		
			rx_count += res;
			if (rx_count > MaxRxMsgLength)
			{
				logit("e","[ERROR] Rx buffer overflow, received %d bytes while MaxRxMsgLength is %d, increase buffer size or check tty bitrate!!!\r\n", rx_count, MaxRxMsgLength);
				exit (-1);
			}
			memcpy(RxBuffer_ptr,data_buff_ptr, res);
			
			RxBuffer_ptr += res;
			
		}
		
		
		
		
		// got timeout
		if (rx_count > 0)
		{
		
			
			//printf("checking for incoming data (%d bytes)...\r\n", rx_count);
			
			RxBuffer_ptr = RxBuffer_ptr_limit = RxBuffer;	// init pointers
			RxBuffer_ptr_limit += rx_count;
			// resetting all flags
			SystemStatus.flag.data_MOD = 0;
			SystemStatus.flag.data_MDE = 0;
			SystemStatus.flag.data_SUM = 0;
			SystemStatus.flag.data_DATA = 0;
			SystemStatus.flag.MsgValid = 1; // init il flag
			
			

			
			do
			{
				// first block should be MOD packet
				
				if (!strcmp((const char *)"MOD", (char *)RxBuffer_ptr))
				{
					// we found MOD packet
					
					SystemStatus.flag.data_MOD = 1;		// set flag for MOD packet
					memcpy(&preo, RxBuffer_ptr+4, 4);	// let's get data buffer length
					DataLen_B = (unsigned short)preo.val;
					DataLen_B += 8;
					if (gcfg_debug > 0)
					{
						printf("found MOD packet (%d bytes) ...\r\n", DataLen_B);
					}
					memcpy(&block_MOD, RxBuffer_ptr, DataLen_B);
					printMOD_ptr = RxBuffer_ptr;	// copy of ptr
					printMOD_n = DataLen_B;
					
					
					RxBuffer_ptr += DataLen_B;	// moving to next block
				}
				else if (!strcmp((const char *)"MDE", (char *)RxBuffer_ptr))
				{
					// we found MDE packet
					
					SystemStatus.flag.data_MDE = 1;		// set flag for MOD packet
					memcpy(&preo, RxBuffer_ptr+4, 4);	// let's get data buffer length
					DataLen_B = (unsigned short)preo.val;
					DataLen_B += 8;
					if (gcfg_debug > 0)
					{		
						printf("found MDE packet (%d bytes)...\r\n", DataLen_B);
					}
					memcpy(&block_MDE, RxBuffer_ptr, DataLen_B);
					RxBuffer_ptr += DataLen_B;	// moving to next block					
				}
				else if (!strcmp((const char *)"DAT", (char *)RxBuffer_ptr))
				{
					// we found DATA packet
					if (gcfg_debug > 0)
					{
						printf("found DATA packet...\r\n");
					}
					DATA_ptr = (unsigned char *)RxBuffer_ptr; 
					DATA_ptr += 8; // ptr to data area 
					SystemStatus.flag.data_DATA = 1;		// set flag for MOD packet
					memcpy(&preo, RxBuffer_ptr+4, 4);	// let's get data buffer length
					DataLen_B = (unsigned short)preo.val;
					DataLen_B += 8;
					memcpy(&block_DATA, RxBuffer_ptr, DataLen_B);
					RxBuffer_ptr += DataLen_B;	// moving to next block				
				}
				else if (!strcmp((const char *)"SUM", (char *)RxBuffer_ptr))
				{
					// we found SUM packet
					if (gcfg_debug > 0)
					{
						printf("found SUM packet...\r\n");
					}
					SUM_ptr = (unsigned char *)RxBuffer_ptr;		// get pointer copy for checksum (later)
					SystemStatus.flag.data_SUM = 1;		// set flag for MOD packet
					memcpy(&preo, RxBuffer_ptr+4, 4);	// let's get data buffer length
					DataLen_B = (unsigned short)preo.val;
					DataLen_B += 8;
					
					memcpy(&block_SUM, RxBuffer_ptr, DataLen_B);
					RxBuffer_ptr += DataLen_B;	// moving to next block				
				}		
				else
				{
					// allowed blocks are only MOD, MDE, DAT, SUM
//					printf("unknown block type....\r\n");
					logit("et","[Warning] received unknown block type...\r\n");
					// resetting all flags, discarding all data
					SystemStatus.flag.MsgValid = 0; // reset flag
					SystemStatus.flag.data_MOD = 0;
					SystemStatus.flag.data_MDE = 0;
					SystemStatus.flag.data_DATA = 0;
					SystemStatus.flag.data_SUM = 0;

				}
			} while ((RxBuffer_ptr < RxBuffer_ptr_limit) && (SystemStatus.flag.MsgValid));
			
		
			if ((SystemStatus.flag.data_SUM) && (SystemStatus.flag.MsgValid))
			{
				if (gcfg_debug > 2)
				{
					printf("checksum...\r\n");
				}
				// found SUM packet, let's check for checksum....
				SUM_ptr += 10;
				
				
	 			
				
				RxBuffer_ptr = RxBuffer;	// resetting pointer
				
				unsigned short checksum, chkcmp;
				checksum = 0;
				u = 0;
				while((unsigned char *)RxBuffer_ptr < SUM_ptr)
				{
						//printf("%04X +%02X\r\n",(unsigned int)(RxBuffer_ptr-RxPacket),  *RxBuffer_ptr);
					// keep in mind checksum needs unsigned char!!!!
					checksum += (unsigned char)*RxBuffer_ptr;
					
					RxBuffer_ptr++;
					u++;
				}
			
				
				
				chkcmp = block_SUM.ChecksumHi;
				chkcmp = chkcmp << 8;
				chkcmp |=  block_SUM.ChecksumLo;
				
				 SystemStatus.flag.ChkPass = 0;  // init flag
				if (chkcmp == checksum)
				{
					SystemStatus.flag.ChkPass = 1;
					if (gcfg_debug > 1)
					{
						printf("\tpass!\r\n");
					}
				}
				else
				{
					if (gcfg_debug > 1)
					{
						printf("\tFAIL!\r\n");
					}
				}
				SystemStatus.flag.data_SUM = 0; // resetting flag
			
			}
			
			
			if ((SystemStatus.flag.ChkPass) && (SystemStatus.flag.data_MOD))
			{
				SystemStatus.flag.ChkPass = 0;	//resetting flag
				// checksum pass, let's get data
				
				
				// let's check if first time we receive message, so let's get digitizer configuration from MOD packet
				if (!SystemStatus.flag.ConfigInit)
				{
					SystemStatus.flag.ConfigInit = 1;
					
					// i blocchi ASCII non vengono chiusi con le stringhe ma con 0x20, io inserisco il carattere NULL
					block_MOD.DeviceID[11] = 0x00;
					block_MOD.Space_1 = 0x00;
					block_MOD.Space_2 = 0x00;
					
					
					logit("","found device: %s\r\n", block_MOD.DeviceID);
					logit("","        ver.: %s\r\n", block_MOD.Version);
                    logit("","         S/N: #%s\r\n", block_MOD.serialNo);
					


					DeviceConfig.Ncomps = block_MOD.Ncomps; 				// get number of components
					logit("","    channels: %d\r\n", block_MOD.Ncomps);
					
					
					
					
					DeviceConfig.sample_rate = block_MOD.sample_rate;       // get total samples of current block
							
					DeviceConfig.Bytesper = block_MOD.Bytesper;     		// get bytes/sample
					DeviceConfig.Gain = block_MOD.Gain;     				// get channel gain
					DeviceConfig.Cal_gain_C1 = block_MOD.Cal_gain_C1;       // get number of components
					DeviceConfig.Cal_gain_C2 = block_MOD.Cal_gain_C2;       // get number of components
					DeviceConfig.Cal_gain_C3 = block_MOD.Cal_gain_C3;       // get number of components
					NofC = DeviceConfig.Ncomps;								// number of channels per block
					NofB = DeviceConfig.Bytesper;							// number of bytes/sample
					
					
					if (NofC > Max_nSCNL)
					{
						logit("e","[ERROR] detected %d channels in block data while allowed up to %d channels, please increase Max_nSCNL in edata2ew.h!\r\n", NofC, Max_nSCNL);
						exit (0);
					}
					if (NofC > nSCNL)
					{
						logit("e","[ERROR] detected %d channels in block data but only following %d channels are defined in configuration file:\r\n", NofC, nSCNL);
						for (u=0; u<nSCNL; u++)
						{
							logit("","\tch%d = %s.%s\r\n", (u+1), Traces.trace[u].cmp,  Traces.trace[u].loc); 
						}
						exit (0);
					}
					if (NofC < nSCNL)
					{
						logit("e","[ERROR] detected only %d channels in block data while following channels are defined in configuration file, please check configuration file!\r\n", NofC);
						for (u=0; u<nSCNL; u++)
						{
							logit("","\tch%d = %s.%s\r\n", (u+1), Traces.trace[u].cmp,  Traces.trace[u].loc); 
						}
						
						exit (0);
					}



					
					
					Fs = DeviceConfig.sample_rate/NofC;						// sampling frequency [SPS]
					logit("","    sampling rate: %d SPS (%d bytes/sample)\r\n", Fs, NofB);			
					dT = 1.0/(double)Fs;									// sampling period [s]

					//wf.trh2.samprate = (double)Fs; 
					DummyTrace.header.samprate = (double)Fs;	// updating header 


					if (gcfg_debug > 0)
					{
						printf("\t\t\tcomponents: %d\r\n", block_MOD.Ncomps);
						printf("\t\t\t%d SPS, %d bytes/sample\r\n", Fs, block_MOD.Bytesper);
						/*printf("\t\t\tgain %d\r\n", block_MOD.Gain);
						printf("\t\t\tcalibration gain:\r\n");

						printf("\t\t\t\tch1 = %ld\r\n", block_MOD.Cal_gain_C1);
						printf("\t\t\t\tch2 = %ld\r\n", block_MOD.Cal_gain_C2);
						printf("\t\t\t\tch3 = %ld\r\n", block_MOD.Cal_gain_C3);

						printf("\t\t\toffset:\r\n");

						printf("\t\t\t\tch1 = %ld\r\n", block_MOD.Offset_C1);
						printf("\t\t\t\tch2 = %ld\r\n", block_MOD.Offset_C2);
						printf("\t\t\t\tch3 = %ld\r\n", block_MOD.Offset_C3);
						*/
					}
				}
				else
				{
					// digitizer configuration can be changed via modmon32 via serial, I don't know if settings can be changed via web interface
					// so let's suppose I can receive MOD packed with updated settings
					
					if (DeviceConfig.Ncomps != block_MOD.Ncomps)
					{
						logit("e","[ERROR] number of channels in data block has changed, please update configuration file!\r\n");
						exit (0);
					}		
						
						
					if ( (DeviceConfig.sample_rate != block_MOD.sample_rate) || (DeviceConfig.Bytesper != block_MOD.Bytesper))
					{
						// let's update configuration
						
						
					
						
						
						logit("","Detected new digitizer settings, updating...\r\n");
						

						
						
								
						DeviceConfig.sample_rate = block_MOD.sample_rate;       // get total samples of current block							
						DeviceConfig.Bytesper = block_MOD.Bytesper;     		// get bytes/sample
					
						NofC = DeviceConfig.Ncomps;								// number of channels per block
						NofB = DeviceConfig.Bytesper;							// number of bytes/sample
						Fs = DeviceConfig.sample_rate/NofC;						// sampling frequency [SPS]
								
						dT = 1.0/(double)Fs;									// sampling period [s]
						DummyTrace.header.samprate = (double)Fs;				// updating header 
					}
				}
				
				
				
				
				
				
				if (gcfg_debug > 4)
				{
					printf("MOD block #%lu\r\n",  block_MOD.block_count);
				}
				
				
				if (gcfg_debug > 5)
				{
					printHexData((unsigned char *)printMOD_ptr, printMOD_n);
						
				}
				
				
				if (gcfg_debug > 2)
				{
					#ifdef EMULATED_GPRMC	
						printf("block timestamp (from MOD.ltime): 0 seconds\r\n");
					#else
						printf("block timestamp (from MOD.ltime): %lu seconds\r\n", block_MOD.ltime);
					#endif
					
					printf("PLL value: %d\r\n", block_MOD.Plldata);

				}
				
				
				
				
				
#ifdef EMULATED_GPRMC			
				// convert timestamp 
				time_ltime = (time_t)block_MOD.ltime;
				//printf("%lu seconds for NMEA string...\r\n", time_ltime);
				
				tm_NMEA = gmtime(&time_ltime);
				// adesso vado a creare il mio messaggio NMEA
				sprintf(MyNMEA,"$GPRMC,%02d%02d%02d,A,3751.65,S,14507.36,E,000.0,360.0,%02d%02d%02d,011.3,E*", tm_NMEA->tm_hour, tm_NMEA->tm_min, tm_NMEA->tm_sec, tm_NMEA->tm_mday, tm_NMEA->tm_mon+1, (tm_NMEA->tm_year+1900)%100);
				
				GPRMC_emulation(MyNMEA);
				//printf("%s\r\n", MyNMEA);
				//printf("%d\r\n", strlen(MyNMEA));
				// clear GPS_message
				for (unsigned char w=0; w<129; w++)
				{
					block_MOD.GPS_message[w] = 0;
				}
				
				memcpy(&block_MOD.GPS_message,&MyNMEA, strlen(MyNMEA));
				
				block_MOD.ltime = 0;
				//printf("clearing block_MOD.ltime...\r\n"); 

#endif				
				
				// let's check if available ltime, if not let's read GPS message for timestamp....
				if ((block_MOD.ltime == 0) && (block_MOD.block_count != 0))
				{
					epoch = 0;	// init timestamp
					// ok, it could be timestamp is not present in ltime field but inside $GPRMC string
					if (block_MOD.GPS_message[0] == '$')
					{
						//printf("detected NMEA GPS string...\r\n");
						// ok, let's read data from GPS message
						if (GPRMC_parsing(block_MOD.GPS_message) == 0)
						{
							//printf("got time from $GPRMC string...\r\n");
							// ok, got timestamp from NMEA string....
							block_MOD.ltime = epoch;
							if (gcfg_debug > 2)
							{
								printf("block timestamp (from MOD.GPS_message): %lu seconds\r\n", block_MOD.ltime);								
							}
						}
						
						
						
					}						
				}	
				
				
				
				
				
				
				
				
				
				
				
				
				
				
				
				// week rollover check, if detected GPS time within 1 Oct 2001 and 1 Jan 2021, probably there is a GPS chip with week rollover bug
				//dummy_double = block_MOD.ltime;
				
				// 1 Jan 2069
//				if (block_MOD.ltime > 3124224000)
//				{
//					// when no GPS antenna connected I get this timestamp...
//				}
				
				
				if (gcfg_WNRO)
				{
					// enabled WNRO offset
					if ((block_MOD.ltime > 1001894400) && (block_MOD.ltime<1609459200))
					{
						// probably WNRO....
						block_MOD.ltime += WNRO_offset;	// year rollover
						if (gcfg_debug > 2)
						{
							printf("updated time: %lu seconds\r\n", block_MOD.ltime);
						}
					}
				}	
						
				dummy_double = block_MOD.ltime;
						
						
				
				if (gcfg_debug > 2)
				{					
					strDate[0] =0;	
					//date23(dummy_double+GSEC1970, strDate);
					date20(dummy_double+GSEC1970, strDate);
					printf("%s UTC\r\n", strDate);
				}
				
				// if found DAT packet, if found MOD packet and got sampling frequency value, let's read data from digitizer
				if (SystemStatus.flag.data_DATA && Fs && SystemStatus.flag.data_MOD)
				{
					unsigned short 	u;
					unsigned short 	nsamp;
					unsigned char 	offB, shiB;
					unsigned int 	msg_size;
					u_Sample_i4 	SampleValue;	// conversione byte -> int
					
					
					
					if (block_DATA.SizeofDAT == (block_MOD.sample_rate*NofB))
					{
						if (gcfg_debug > 2)
						{
							printf("%ld data bytes on current block (@%d samples/block, %dSPS)...\r\n",  block_DATA.SizeofDAT, DeviceConfig.sample_rate, Fs);
						}
						// init local buffer
						Traces.npts = 0;
						Traces.starttime = block_MOD.ltime;
						
						
						offB = 4-NofB;	// offset di partenza di SampleValue.v
						shiB = 8*offB;	// shif right
						for (nsamp = 0; nsamp < Fs; nsamp++)
						{
							for (unsigned char cmp=0; cmp<NofC; cmp++)
							{
								// voglio riportare direttamente il segno
								// NofB = 4 -> parto con u a 0
								// NofB = 3 -> u = 1
								// NofB = 2 -> u = 2
								// NofB = 1 -> u = 3
								SampleValue.val = 0;
								for (u=offB; u<4; u++)
								{
									//printf("reading %02X -> .v[%d]\r\n", (unsigned int) *DATA_ptr & 0xFF, u);
									SampleValue.v[u] = *DATA_ptr++;	// conversione byte -> int
								}
								// adesso devo shiftare di tanti bytes
								// 4 -> 0b
								// 3 -> 1b
								if (NofB != 4)
								{
									SampleValue.val = SampleValue.val >> shiB;
								}
								// quindi ho riempito i primi NoB bytes di SampleValue
								// con NofB = 4, qui mi ritrovo u=4
								// con NofB = 3, qui mi ritrovo u=3
								// con NofB = 2, qui mi ritrovo u=2
								// con NofB = 1, qui mi ritrovo u=1
								// adesso devo riportare il segno
								Traces.trace[cmp].data[nsamp] = SampleValue.val;
							}
							Traces.npts++;	
						}
						
						// let's check if something wrong during reading...
						if (Traces.npts == nsamp)
						{
							// ok, matching counters
							
							
							//printf("Got stream (%d) [%.06f --> %06f]\r\n", nsamp, (double)Traces.starttime, (double)Traces.starttime)
							for (unsigned char cmp=0; cmp<NofC; cmp++)
							{
								
								
								
								
								memcpy(DummyTrace.header.chan, Traces.trace[cmp].cmp, 4);
								memcpy(DummyTrace.header.loc, Traces.trace[cmp].loc, 3);



								DummyTrace.header.starttime = DummyTrace.header.endtime = (double)Traces.starttime;
								DummyTrace.header.nsamp = nsamp;
								DummyTrace.header.endtime += dT*(nsamp-1); 
								
								twrite_ptr = DummyTrace.databuffer;						// pointer to dst buffer 
								tread_ptr = (char *)Traces.trace[cmp].data;		// pointer to src buffer
								
								msg_size = 4*nsamp;										// 1 sample in local buffer = 4 bytes
								
								memcpy(twrite_ptr, tread_ptr, msg_size);				// copy data from local buffer
								
								
								
								msg_size += sizeof(TRACE2_HEADER);	// incremento con il numero di bytes dell'header
								
								if (gcfg_debug > 1)
								{								
									printf("sending trace %s.%s.%s.%s (%d) [%.03f --> %.03f]\r\n", DummyTrace.header.sta, DummyTrace.header.chan, DummyTrace.header.net, DummyTrace.header.loc, DummyTrace.header.nsamp, DummyTrace.header.starttime, DummyTrace.header.endtime);		
								}	
									
#ifndef	standalone
								if ( tport_putmsg( &Region, &SendMsg, msg_size, (char*)&DummyTrace ) != PUT_OK )
								{
									logit("et", "edata2ew: Error sending message via transport.\n");
									return EW_FAILURE;
								}
								else
								{
									if (gcfg_debug > 3)
									{
										printf("\t...success!\r\n");
									}
								}
#endif


							}
							Traces.npts = 0;
							
						}
						
						
						
					}
					else
					{
							logit("e","[ERROR] mismatching samples on data block.\r\n");
					}
					
				}
				
				
				
			}
			
			
			
			
			
		}
		rx_count = 0;
		
		
	}
	close(ttyfd);

//-----------------------------end of main loop-------------------------------*/        
	
}

#define EDATA_MAX_STREAMS 6
#define CHANNEL_ID_LENGTH 3
#define LOCATION_ID_LENGTH 3   /* Length of location name, plus terminal null */

static char channel_names_str[(EDATA_MAX_STREAMS+2)*(CHANNEL_ID_LENGTH+3)];
static char location_names_str[(EDATA_MAX_STREAMS+2)*(CHANNEL_ID_LENGTH+3)];
char *gcfg_chnames_arr[EDATA_MAX_STREAMS];
char *gcfg_locnames_arr[EDATA_MAX_STREAMS];

static char channel_names[EDATA_MAX_STREAMS][4];



/************************************************************************
 * edata_parse_chnames_str:  Parses channel-name items from the given string; *
 *      items may be separated by spaces or commas; the global          *
 *      'gcfg_chnames_arr[]' array is loaded with pointers to           *
 *      the items found in the given string; returns 0 if successful,   *
 *      returns -1 if error.                                            *
 ************************************************************************/
int edata_parse_chnames_str(char *chnames_str)	
{
	
	unsigned char i, chn, lc;
	char *ptr;
	char lostr[4];
	char maxchar;
	
	
	
	//printf("input string <%s>\r\n", chnames_str);
	
	i = 0;
	chn = 0;
	ptr = chnames_str;
	lc = 0;
	maxchar = 4;
	while ((*ptr != '#') && (*ptr != '\r') && (*ptr != '\n') && (*ptr != 0x00))
	{
		//printf("checking char %c at offset %d\r\n", *ptr, i);
		if (((*ptr > 0x40) && (*ptr < 0x5B)) || ((*ptr > 0x60) && (*ptr < 0x7B))  || ((*ptr > 0x2f) && (*ptr < 0x3A)))
		{
			//printf(" - allowed char within ranges\r\n");
			if ((*ptr > 0x60) && (*ptr < 0x7B))
			{
				*ptr -= 32;	// convert to uppercase
			}
			lostr[lc] = *ptr;
			lc++;
			if (lc == maxchar)
			{
				printf("[error] maxsize\r\n");
				// channel code higher than expected
				return (-1);
			}
		}

		if ((*ptr == ' ') || (*ptr == ',') || (*ptr == ';')  )
		{
			//printf(" - detected separator\r\n");
			if (lc == (maxchar-1))
			{
				
				if (chn == EDATA_MAX_STREAMS)
				{
					logit("e","[ERROR] allowed max %d streams/channels, check configuration file!\r\n", EDATA_MAX_STREAMS);
					return (-2);
				}
				lostr[lc] = 0x00;
				
				memcpy(Traces.trace[chn].cmp,lostr,maxchar); 
				
				chn++;
				//printf("#%d: %s\r\n", chn, lostr);
				
				// re-init
				lc = 0;	
			}
		}
		i++;
		ptr++;
	}
	//printf("checking char %c at offset %d\r\n", *ptr, i);
	if ((*ptr == '#') || (*ptr == '\r') || (*ptr == '\n') || (*ptr == 0x00))
	{
		
		//printf(" - detected EOC\r\n");
		// detected end command
		if (lc == (maxchar-1))
		{
			
			if (chn == EDATA_MAX_STREAMS)
			{
				logit("e","[ERROR] allowed max %d streams/channels, check configuration file!\r\n", EDATA_MAX_STREAMS);
				return (-3);
			}
			lostr[lc] = 0x00;
			memcpy(Traces.trace[chn].cmp,lostr,maxchar); 
			chn++;
			//printf("#%d: %s\r\n", chn, lostr);
		}
	}
	
	// init location to default "--" value
	for (i=0; i<chn; i++)
	{

		Traces.trace[i].loc[0]=0; //, "--");
		sprintf(Traces.trace[i].loc, "--");
		

	}
	

	
	nSCNL = chn;

	return 0;        /* return OK value */	
}


int edata_parse_locames_str(char *locames_str)	
{
	
	unsigned char i, chn, lc;
	char *ptr;
	char lostr[4];
	char maxchar;
	
	
	if (nSCNL == 0)
	{
		// I need channel names to be set before location names
		logit("e","[Warning] LocationNames command placed before ChannelNames, location set to default value\r\n");
		return 0; 
	}
	
	//printf("input string <%s>\r\n", chnames_str);
	
	i = 0;
	chn = 0;
	ptr = locames_str;
	lc = 0;
	maxchar = 3;
	while ((*ptr != '#') && (*ptr != '\r') && (*ptr != '\n') && (*ptr != 0x00))
	{
		//printf("checking char %c at offset %d\r\n", *ptr, i);
		if (((*ptr > 0x40) && (*ptr < 0x5B)) || ((*ptr > 0x60) && (*ptr < 0x7B))  || (*ptr == '-') || ((*ptr > 0x2f) && (*ptr < 0x3A)))
		{
			//printf(" - allowed char within ranges\r\n");
			if ((*ptr > 0x60) && (*ptr < 0x7B))
			{
				*ptr -= 32;	// convert to uppercase
			}
			lostr[lc] = *ptr;
			lc++;
			if (lc == maxchar)
			{
				//printf("[error] maxsize\r\n");
				// channel code higher than expected
				return (-1);
			}
		}

		if ((*ptr == ' ') || (*ptr == ',') || (*ptr == ';')  )
		{
			//printf(" - detected separator\r\n");
			if (lc == (maxchar-1))
			{
				
				if (chn == nSCNL)
				{
					logit("e","[ERROR] detected more location names code than allowed channels!\r\n");
					return (-2);
				}
				lostr[lc] = 0x00;
				
				memcpy(Traces.trace[chn].loc,lostr,maxchar); 
				chn++;
				//printf("#%d: %s\r\n", chn, lostr);
				
				// re-init
				lc = 0;	
			}
		}
		i++;
		ptr++;
	}
	//printf("checking char %c at offset %d\r\n", *ptr, i);
	if ((*ptr == '#') || (*ptr == '\r') || (*ptr == '\n') || (*ptr == 0x00))
	{
		
		//printf(" - detected EOC\r\n");
		// detected end command
		if (lc == (maxchar-1))
		{
			
			if (chn == nSCNL)
			{
				logit("e","[ERROR] detected more location names code than allowed channels!\r\n");
				return (-3);
			}
			lostr[lc] = 0x00;
			memcpy(Traces.trace[chn].loc,lostr,maxchar); 
			chn++;
			//printf("#%d: %s\r\n", chn, lostr);
		}
	}
	
	
	if (chn < nSCNL)
	{
		logit("e","[warning] Detected %d channels but only %d locations, filling remaining locations to default values...\r\n", nSCNL, chn);
		for (i=chn; i<nSCNL; i++)
		{
			sprintf(Traces.trace[i].loc, "--");
		}
		
	}

	
	

	return 0;        /* return OK value */	
}







/******************************************************************************
 *  edata2ew_ReadConfigFile() processes command file(s) using kom.c functions;        *
 *                    exits if any errors are encountered.                    *
 ******************************************************************************/
int edata2ew_ReadConfigFile( char *configfile )
{
   int      ncommand;     // # of required commands you expect to process    
   char     init[10];     // init flags, one byte for each required command 
   int      nmiss;        // number of required commands that were missed   
   char    *com;
   char    *str;
   int      nfiles;
   int      success;
   int      i;
   char     *sptr,*sptr2;
   int len;

	// Set to zero one init flag for each required command 
  
   ncommand = 5;
   for( i=0; i<ncommand; i++ )  
   {
	   init[i] = 0;
   }
   nLogo = 0;

   for(i=0; i<EDATA_MAX_STREAMS; ++i)
   {     /* for each channel (or stream) */
     gcfg_chnames_arr[i] = NULL;
     gcfg_locnames_arr[i] = NULL;
     
   }


	for (i=0; i<EDATA_MAX_STREAMS; i++)
	{
		for (len=0; len<4 ; len++)
		{
			channel_names[i][len]=0x00;
		}
	}




	// Open the main configuration file 
   nfiles = k_open( configfile ); 
   if ( nfiles == 0 ) 
   {
        logit( "e", "edata2ew: Error opening command file <%s>; exiting!\n", configfile );
        exit( -1 );
   }

	// Process all command files
	while(nfiles > 0)   /* While there are command files open */
	{
        while(k_rd())        /* Read next line from active file  */
        {  
			com = k_str();         /* Get the first token from line */

			// Ignore blank lines & comments
            if( !com )
			{
				continue;
			}
            if( com[0] == '#' )  
			{
				continue;
			}
			
			// Open a nested configuration file 
            if( com[0] == '@' ) 
			{
				success = nfiles+1;
				nfiles  = k_open(&com[1]);
				if ( nfiles != success ) 
				{
					logit( "e", "edata2ew: Error opening command file <%s>; exiting!\n", &com[1] );
					exit( -1 );
				}
				continue;
            }

			// Process anything else as a command 
			if( k_its("LogFile") ) 							// 0
			{
                LogSwitch = k_int();
                init[0] = 1;
            }
			else if( k_its("ModuleId") ) 
			{
                str = k_str();
                if(str) strcpy( MyModName, str );
                init[1] = 1;
            }
			else if( k_its("RingName") ) 
			{
                str = k_str();
                if(str) 
				{
					strcpy( RingName, str );
				}
                init[2] = 1;
            }
			else if( k_its("HeartbeatInterval") ) 
			{
                HeartBeatInterval = k_long();
                init[3] = 1;
            }

			else if( k_its("TtyName") ) 
			{
				
				if ( (str = k_str()) != 0 )
				{
					if (strlen(str) >= MAX_TTY_LEN - 1)
					{
						logit("e","check TtyName: is longer than allowed (%d chars)\r\n", MAX_TTY_LEN);
						return (-1);
					}
					strcpy(SystemConfig.tty.name, str);
				}
				init[4] = 1;
			}
		 
			
		 
			else if ( k_its( "Speed" ) )
			{
				if((SystemConfig.tty.bitrate=k_long()) < 150L || SystemConfig.tty.bitrate > 10000000L)
				{
					logit("e","check Speed: value not within allowed range.\r\n");
					return (-1); 
				}				
				init[5] = 1;
			}

			else if ( k_its( "StationCode" ) )
			{ 
				// copy station code name; make sure NULL terminated 
				str = k_str();
				if (strlen(str) > 5)
				{
					logit("e","check StationCode: name is longer than allowed (5 chars)\r\n");
					return (-1);
				}
				if(str) 
				{
					strcpy( SystemConfig.StationCode, str );
					init[6] = 1;
				}	
			}
			else if ( k_its( "ChannelNames" ) )
			{
				
				if ((str = k_str()) != NULL && strlen(str) > 0)
				{
					
					edata_parse_chnames_str(str);
					init[7] = 1;
				}
			}



			else if( k_its("Network") ) 
			{
				str = k_str();
				
				if ((str != NULL) && (strlen(str) == 2))
				{
					
					SystemConfig.NetworkName[2] = 0;
				
					strncpy( SystemConfig.NetworkName, str, 2 );
				
					init[8] = 1;
				}
				else
				{
					logit("e","Error reading network code\r\n");
					return -1;
				}
				
				
				
				
			}



			else if ( k_its( "LocationNames" ) )
			{
				if ((str = k_str()) != NULL && strlen(str) > 0)
				{
					
					edata_parse_locames_str(str);
					
				}
			} // fine else if ( k_its( "LocationNames" ) )

			else if ( k_its( "Debug" ) )
			{
				
                gcfg_debug = k_long();
			}
			else if ( k_its( "WNRO_offset" ) )
			{
				
                gcfg_WNRO = k_long();
			}
			








			//	 Unknown command
            else {
                logit( "e", "edata2ew: <%s> Unknown command in <%s>.\n", 
                         com, configfile );
                continue;
            }

        /* See if there were any errors processing the command 
         *****************************************************/
            if( k_err() ) {
               logit( "e", 
                       "edata2ew: Bad <%s> command in <%s>; exiting!\n",
                        com, configfile );
               exit( -1 );
            }
        }
        nfiles = k_close();
   }

// After all files are closed, check init flags for missed commands
   nmiss = 0;
   for ( i=0; i<ncommand; i++ )  
   {
	   if( !init[i] ) 
		   nmiss++;
   }
   if ( nmiss ) {
       logit( "e", "edata2ew: ERROR, no " );
       if ( !init[0] )  logit( "e", "<LogFile> "           );
       if ( !init[1] )  logit( "e", "<MyModuleId> "        );
       if ( !init[2] )  logit( "e", "<RingName> "          );
       if ( !init[3] )  logit( "e", "<HeartBeatInterval> " );
       //if ( !init[4] )  logit( "e", "<GetEventsFrom> "     );
       logit( "e", "command(s) in <%s>; exiting!\n", configfile );
       exit( -1 );
   }


/*

	for (i=0; i<20; i++)
	{
		printf("%d <%c>[%02X]\r\n", i, channel_names_str[i], channel_names_str[i]);
	}

	for (i=0; i<20; i++)
	{
		printf("%d <%c>[%02X]\r\n", i, location_names_str[i], location_names_str[i]);
	}


*/




   return 0;
}

/******************************************************************************
 *  edata2ew_lookup( )   Look up important info from earthworm*.d tables       *
 ******************************************************************************/
void edata2ew_lookup( void )
{
/* Look up keys to shared memory regions
   *************************************/
   if( ( RingKey = GetKey(RingName) ) == -1 ) 
   {
        fprintf( stderr, "edata2ew:  Invalid ring name, check earthworm.d for <%s> ring; exiting!\n", RingName);
        exit( -1 );
   }

/* Look up installations of interest
   *********************************/
   if ( GetLocalInst( &InstId ) != 0 ) 
   {
      fprintf( stderr, "edata2ew: error getting local installation id; exiting!\n" );
      exit( -1 );
   }

/* Look up modules of interest
   ***************************/
   if ( GetModId( MyModName, &MyModId ) != 0 ) 
   {
      fprintf( stderr, "edata2ew: Invalid module name, check earthworm.d for <%s> module; exiting!\n", MyModName );
      exit( -1 );
   }

/* Look up message types of interest
   *********************************/
   if ( GetType( "TYPE_HEARTBEAT", &TypeHeartBeat ) != 0 ) 
   {
      fprintf( stderr, 
              "edata2ew: Invalid message type <TYPE_HEARTBEAT>; exiting!\n" );
      exit( -1 );
   }
	if ( GetType( "TYPE_TRACEBUF2", &TypeTrace2 ) != 0 ) 
	{
		 fprintf( stderr, 
              "edata2ew: Invalid message type <TYPE_TRACEBUF2>; exiting!\n" );
		exit( -1 );
	}   
   if ( GetType( "TYPE_ERROR", &TypeError ) != 0 ) 
   {
      fprintf( stderr, "edata2ew: Invalid message type <TYPE_ERROR>; exiting!\n" );
      exit( -1 );
   }
  
   return;
} 





















/******************************************************************************
 * edata2ew_status() builds a heartbeat or error message & puts it into       *
 *                   shared memory.  Writes errors to log file & screen.      *
 ******************************************************************************/
void edata2ew_status( unsigned char type, short ierr, char *note )
{
   MSG_LOGO    logo;
   char        msg[256];
   long        size;
   time_t        t;
 
// Build the message

   logo.instid = InstId;
   logo.mod    = MyModId;
   logo.type   = type;

   time( &t );

   if( type == TypeHeartBeat )
   {
        sprintf( msg, "%ld %ld\n\0", (long) t, (long) MyPid);
   }
   else if( type == TypeError )
   {
        sprintf( msg, "%ld %hd %s\n\0", (long) t, ierr, note);
        logit( "et", "edata2ew: %s\n", note );
   }

   size = strlen( msg );   // don't include the null byte in the message     

// Write the message to shared memory

   if( tport_putmsg( &Region, &logo, size, msg ) != PUT_OK )
   {
        if( type == TypeHeartBeat ) {
           logit("et","edata2ew:  Error sending heartbeat.\n" );
        }
        else if( type == TypeError ) {
           logit("et","edata2ew:  Error sending error:%d.\n", ierr );
        }
   }

   return;
}


void  edata2ew_log_cfgparams(void)
{
	
}














// tty functions




/*
int tfd = -1;                   // Our tty port 


// A buffer for recv functions and its counters  
static unsigned char recv_buff[RB_LEN]; 
static int rb_read = 0; 	// Where to start reading from recv_buff  
static int rb_count = 0; 	// Number of bytes in recv_buff; note: we only write 
							// (by calling recv()) into recv_buff when it is empty  
							
							
							
// Internal function prototypes; 
static int GetMoreBytes(int, int);


*/

/************************************************************************
 * k2c_init_io:  tcp port initialization routine                        *
 *      gen_io: generalize IO parameter structure                       *
 *      toutms: timeout in milliseconds; not used                       *
 *      return value:  returns K2R_NO_ERROR if successful               *
 *                     returns EDATAR_ERROR on error                       *
 *                     or K2R_TIMEOUT                                   *
 ************************************************************************/

/*
// qui passo in ingresso una struttura contenente sia il nome del device, sia il bitrate
int edata_init_io(s_tty_io *gen_io, int toutms)
{
	long lOnOff;
	struct termios modes;
	int baud;
  

	// open the TTY device, if it isn't already 
	if (tfd < 0)
		tfd = open( gen_io->ttyname, O_RDWR );
	if ( tfd < 0 ) 
	{
		logit("et", "edata_init_io: error opening TTY port <%s>: %s\n", gen_io->ttyname, strerror(errno));
		return EDATAR_ERROR;
	}

	// Get current settings so we can mess with them 
	if (tcgetattr(tfd, &modes) < 0) 
	{
		close(tfd);
		tfd = -1;
		logit("et", "edata_init_io: get tty attribute error: %s\n", 
			strerror(errno));
		return EDATAR_ERROR;
	}

	// Turn off echo, canonical, extended input, and keyboard signals 
	modes.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
	// Turn off interrupt on break, cr mapping, parity, strip and SW flow control 
	modes.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON | IXOFF);
	// Turn off size mask and parity 
	modes.c_cflag &= ~(CSIZE | PARENB);
	// Set 8-bit size 
	modes.c_cflag |= CS8;
	// Turn off output processing 
	modes.c_oflag &= ~(OPOST);
	// Enable character-at-a-time input 
	modes.c_cc[VMIN] = 1;
	modes.c_cc[VTIME] = 0;
  
	switch(gen_io->speed)
	{
		case 110: baud = B110; break;
		case 150: baud = B150; break;
		case 300: baud = B300; break;
		case 1200: baud = B1200; break;
		case 2400: baud = B2400; break;
		case 4800: baud = B4800; break;
		case 9600: baud = B9600; break;
		case 19200: baud = B19200; break;
		case 38400: baud = B38400; break;	
		case 57600: baud = B57600; break;
		case 115200: baud = B115200; break;
		case 230400: baud = B230400; break;
 
		default:
			close(tfd);
			tfd = -1;
			logit("et", "edata_init_io: bad TTY speed %d\n", gen_io->speed);
			return EDATAR_ERROR;
	}
  
	cfsetospeed(&modes, baud);
	cfsetispeed(&modes, baud);
  
	// Make the actual settings 
	if (tcsetattr(tfd, TCSAFLUSH, &modes) < 0) 
	{
		close(tfd);
		tfd = -1;
		logit("et", "edata_init_io: error setting tty attributes: %s\n", 
          strerror(errno));
		return EDATAR_ERROR;
	}
  
  
	//sleep(2); //required to make flush work, for some reason
	//tcflush(tfd,TCIOFLUSH);
  
  
	// Now set the tty to non-blocking mode 
	lOnOff = 1;  // on 
	if (ioctl(tfd, FIONBIO, &lOnOff) < 0)
	{
		close(tfd);
		tfd = -1;
		logit("et", "edata_init_io: ioctl error: %s\n", strerror(errno));
		return EDATAR_ERROR;
	}

	return K2R_NO_ERROR;
}

*/



/************************************************************************
 * k2c_close_io:  closes tty port opened by 'k2c_init_io()'             *
 *         return value: returns K2R_NO_ERROR on success,               *
 *                        K2R_ERROR on error                            *
 ************************************************************************/
/*
int edata_close_io( void )
{
  int rc = K2R_NO_ERROR;
  
  if ( tfd > 0 )
    if ( close( tfd ) < 0)
    {
      logit("et", "edata_close_io: error closing tty (%d): %s\n", tfd, 
            strerror(errno));
      rc = EDATAR_ERROR;
    }
  
  tfd = -1;   // Mark the tty as closed 
  return rc;
}

*/


/************************************************************************
 * GetMoreBytes: an internal function for filling the static recv_buff. *
 *           to_read: number of bytes to recv; rb_count is set to       *
 *               number of bytes actually read; rb_read is set to 0     *
 *            toutms: if non-zero then a timeout value in milliseconds; *
 *                    if zero then no timeout is used.                  *
 *           returns: K2R_NO_ERROR on success, K2R_ERROR on error       *
 *                    or K2R_TIMEOUT on timeout                         *
 ************************************************************************/
/*

// metto i dati in recv_buff[RB_LEN] con RB_LEN=256 e aggiorno rb_read con il valore dei bytes letti
static int GetMoreBytes(int to_read, int toutms)
{
  int n_read, sel_ret;
  fd_set r_set;
  struct timeval tv, *ptv;
  
  if (gcfg_debug > 3)
    logit("et", "enter GMB TO %d\n", toutms);
  
  if (to_read < 1 || to_read > RB_LEN)
  {
    logit("et", "GetMoreBytes: invalid read request <%d>\n", to_read);
    return K2R_ERROR;
  }
  rb_read = rb_count = 0;
  
  if ( (n_read = read(tfd, (char *)recv_buff, to_read)) <= 0)
  {
    if ( n_read < 0 && errno != EAGAIN)
    {
      rb_count = 0;
      logit("et", "GetMoreBytes: read error: %s\n", strerror(errno));
      return K2R_ERROR;
    }
    else
    {  // Would have blocked, so we will select... 
      ptv = &tv;
      tv.tv_sec = toutms / 1000;
      tv.tv_usec = (toutms % 1000) * 1000;
      
      FD_ZERO( &r_set );
      FD_SET(tfd, &r_set);
      
      if (gcfg_debug > 3)
        logit("et", "GMB selecting %lu.%lu\n", tv.tv_sec, tv.tv_usec);
      
      sel_ret = select( tfd+1, &r_set, NULL, NULL, ptv);
      
      if (gcfg_debug > 3)
        logit("et", "GMB sel ret %d\n", sel_ret);
      
      if (sel_ret > 0 && FD_ISSET( tfd, &r_set ))
      {   // Maybe there's something to receive 
        if ( (n_read = read(tfd, (char *)recv_buff, to_read)) < 0)
        {   // We had a another booboo; wrap up and go home 
          logit("et", "GetMoreBytes: read error: %s\n", strerror(errno));
          return K2R_ERROR;
        }
        rb_count = n_read;   // Record how much is in recv_buff 
        if (gcfg_debug > 3)
          logit("e", "GMB: %d\n", rb_count);
        return K2R_NO_ERROR;
      }
      else if (sel_ret == 0)
      {   // Select timed out 
        if (gcfg_debug > 3)
          logit("e", "GMB timeout\n");
        return K2R_TIMEOUT;
      }
      else
      {   // Select had a booboo 
        logit("et", "GetMoreBytes: select error: %s\n", strerror(errno));
        return K2R_ERROR;
      }
    }
    
  }
  // We recv'ed ok the first time 
    rb_count = n_read;
    if (gcfg_debug > 3)
      logit("e", "GMB: %d\n", rb_count);
    return K2R_NO_ERROR;
  }

*/



/**************************************************************************
 * k2c_recv_buff:  fills buffer with received data; waits for given       *
 *      number of bytes to be received; with optional timeout             *
 *         rbuff    - address of buffer to receive data bytes             *
 *         datalen  - number of data bytes to receive                     *
 *         toutms   - timeout value in milliseconds;                      *
 *                                    *
 *      return value: returns the number of data bytes received and       *
 *                    placed into 'rbuff[]'                               *
 *                    returns K2R_ERROR on error                          *
 *                    returns K2R_TIMEOUT if no data read before timeout  *
 **************************************************************************/
/*
int k2c_recv_buff(unsigned char *rbuff, int datalen, int toutms)
{
  int bcount = 0;
  int to_read, to_copy, rc;
  
  if (gcfg_debug > 3)
    logit("e", "k2c_recv_buff %d\n", datalen);
  
  
  while (bcount < datalen)
  {
    if (rb_count <= rb_read)  // Is the recv_buffer empty? 
    {    // Yes, fill it up 
      to_read = (datalen - bcount < RB_LEN) ? datalen - bcount : RB_LEN;
      if ( (rc = GetMoreBytes(to_read, toutms)) != K2R_NO_ERROR)
      {
        if (rc == K2R_ERROR)
          return rc;
        else  // TIMEOUT, return count if we got any 
          return ( (bcount > 0) ? bcount : rc);
      }
    }
    // Copy from recv_buff into our output buffer 
     // Don't exceed capacity of either buffer    
    to_copy = (rb_count - rb_read < datalen - bcount) ? 
      rb_count - rb_read : datalen - bcount;
    memcpy(&rbuff[bcount], &recv_buff[rb_read], to_copy);
    rb_read += to_copy;
    bcount += to_copy;
  }
  return bcount;
}


*/



/**************************************************************************
 * k2c_rcvflg_tout:  waits for data to be received into TTY port          *
 *         toutms   - a timeout value in milliseconds                     *
 *      return value:  returns K2R_NO_ERROR if data is ready to read      *
 *                returns K2R_TIMEOUT if no data to read within timeout   *
 *                     returns K2R_ERROR on error                         *
 **************************************************************************/
/*
int k2c_rcvflg_tout(int toutms)
{
  int sel_ret;
  fd_set r_set;
  struct timeval tv = {0,0};
  
  if (gcfg_debug > 3)
  {
    logit("e", "k2c_rcvflg_tout TO %d\n", toutms);
  }
  FD_ZERO( &r_set );
  FD_SET(tfd, &r_set);
  tv.tv_sec = toutms / 1000;
  tv.tv_usec = (toutms % 1000) * 1000;
  
  sel_ret = select( tfd+1, &r_set, NULL, NULL, &tv);
  if (sel_ret > 0 && FD_ISSET( tfd, &r_set ))
  {   // Looks like there's something to receive 
    if (gcfg_debug > 3)
	{
      logit("e", "k2c_rcvflg_tout ret 0\n");
	}
    return K2R_NO_ERROR;
  }
  else if (sel_ret == 0)
  {   // Select timed out 

    if (gcfg_debug > 3)
	{
      logit("e", "k2c_rcvflg_tout ret TIMEOUT\n");
	}
    return K2R_TIMEOUT;
  }
  else
  {   // Select had a booboo 
    logit("et", "k2c_rcvflg_tout: error in select(): %s\n", strerror(errno));
    return K2R_ERROR;
  }
}


*/


/*



int k2c_flush_recv(int toutms)
{
  int bcount = 0;
  int to_read, rc;
  

  
  rb_read = rb_count = 0;
  if (toutms <= 0)
    bcount = K2C_MAX_FLUSH - 1;
  
  while (bcount < K2C_MAX_FLUSH)
  {
    to_read = (K2C_MAX_FLUSH - bcount < RB_LEN) ? K2C_MAX_FLUSH - bcount : 
      RB_LEN;
    //  if ( (rc = GetMoreBytes(to_read, toutms)) == K2R_ERROR) 
    rc = GetMoreBytes(to_read, toutms);
    if (gcfg_debug > 3)
      logit("e", "GMB ret %d rb %d\n", rc, rb_count);
    if (rc == K2R_ERROR)
      return rc;
    else if (rc == K2R_TIMEOUT)
      return K2R_NO_ERROR;
    
    bcount += rb_count;
  }
  return K2R_NO_ERROR;
}


*/



void edata2ew_PrintConfig (void)
{
	char i;
	if (gcfg_debug > 0)
	{
		printf("Edata2EW configuration:\r\n");

		printf("\tsw. ver.: %s\r\n", MODULE_RELVER_STR);
		printf("\tserial port: %s\r\n", SystemConfig.tty.name);
		printf("\tbitrate: %d [bit/s]\r\n", SystemConfig.tty.bitrate);
		printf("\tstation code: %s\r\n", SystemConfig.StationCode);	
		printf("\tstation network: %s\r\n", SystemConfig.NetworkName);	
		printf("\tdebug level: %d\r\n", gcfg_debug);
	}
	logit("","serial port %s (@%d bit/s) *check sampling frequency and number of channels settings!!!*\r\n",SystemConfig.tty.name, SystemConfig.tty.bitrate);
	logit("","Station: %s.%s\r\n",SystemConfig.NetworkName,  SystemConfig.StationCode);
	for (i=0; i<nSCNL; i++)
	{
		logit("","\tch#%d = %s.%s.%s.%s\r\n", (i+1), SystemConfig.NetworkName,  SystemConfig.StationCode, Traces.trace[i].cmp, Traces.trace[i].loc);
	}
	

	
	
	
	
	
}



char edata_tty_init(char *sPort, unsigned int bitrate)
{
	long lOnOff;
	int baud;


	if (ttyfd < 0)
	{
		ttyfd = open(sPort, O_RDWR );
	}
	
	if (ttyfd < 0)
	{
		printf("Error opening serial port %s: %s\n", sPort, strerror(errno));
		return CODE_ERROR; 
	}

	if (tcgetattr(ttyfd, &tty_options) <0)
	{
		close(ttyfd);
		ttyfd = -1;
		printf("get tty attribute error: %s\n", strerror(errno));
		return CODE_ERROR; 
	}
	
	
	// line options
	// For Raw input: turn off echo, canonical, extended input, and keyboard signals
	tty_options.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);	
	
	//  input options
	// Trun off interrupt on break, cr mapping, parity, strip and flow control (Xon/Xoff disabled)
	tty_options.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXOFF | IXON | IXANY); 
	
	// control options
	tty_options.c_cflag &= ~(CSIZE | PARENB);	// ctob mi mette errore
	
	tty_options.c_cflag |= CS8;	// perchè metto questa? per selezionare la modalità a 8 bit
	
	// output  options
	tty_options.c_oflag &= ~(OPOST);
	// Enable character-at-a-time input 
	tty_options.c_cc[VMIN] = 1;
	tty_options.c_cc[VTIME] = 0;
	
	// adesso imposto il bitrate
	switch(bitrate)
	{
		
		case 19200: baud = B19200; break;
		case 38400: baud = B38400; break;	
		case 57600: baud = B57600; break;
		case 115200: baud = B115200; break;			
		case 230400: baud = B230400; break;		
					
		default:
			close(ttyfd);
			ttyfd = -1;
			printf("Error: bad tty bitrate: %d\n", bitrate);
			return CODE_ERROR;     
	}
	
	cfsetispeed(&tty_options, baud);
	cfsetospeed(&tty_options, baud);
	
	// set the options 
	tcsetattr(ttyfd, TCSAFLUSH, &tty_options);	// ew
	
	
	// Now set the tty to non-blocking mode 
	lOnOff = 1;  /* on */
	ioctl(ttyfd, FIONBIO, &lOnOff); 
	
	// erase previous data inside buffer (useful when using USB-to-serial adapter)
	sleep(1);
	tcflush(ttyfd, TCIOFLUSH);
	return CODE_NO_ERROR;
}




// se non ricordo mi ritorna -1 se sono in timeout, altrimenti il numero di bytes che ho nel buffer
int edata_recv_buff(unsigned char *rbuff, int datalen, int toutms)
{
	int bcount = 0;
	int to_read, to_copy, rc;
	unsigned char gcfg_debug;
	
	gcfg_debug = 0;
	
	
	
	if (gcfg_debug > 3)
	{
		printf("recv_buff %d\n", datalen);
	}
 
	while (bcount < datalen)
	{
		if (rb_count <= rb_read)  /* Is the recv_buffer empty? */
		{    /* Yes, fill it up */
			to_read = (datalen - bcount < RB_LEN) ? datalen - bcount : RB_LEN;
			//printf("to_read = <%d>\r\n", to_read);
			if ( (rc = edata_GetMoreBytes(to_read, toutms)) != K2R_NO_ERROR)
			{
				if (rc == K2R_ERROR)
					return rc;
				else  /* TIMEOUT, return count if we got any */
					return ( (bcount > 0) ? bcount : rc);
			}
		}
		/* Copy from recv_buff into our output buffer *
		* Don't exceed capacity of either buffer     */
		to_copy = (rb_count - rb_read < datalen - bcount) ? rb_count - rb_read : datalen - bcount;
		memcpy(&rbuff[bcount], &recv_buff[rb_read], to_copy);
		rb_read += to_copy;
		bcount += to_copy;
	}
	return bcount;
}




// vado a mettere i dati in recv_buff[RB_LEN]
static int edata_GetMoreBytes(int to_read, int toutms)
{
	int n_read, sel_ret;
	fd_set r_set;
	struct timeval tv, *ptv;
	unsigned char gcfg_debug;
	
	gcfg_debug = 0;
  
  
	if (to_read < 1 || to_read > RB_LEN)
	{
		//printf("GetMoreBytes: invalid read request <%d>\n", to_read);
		return K2R_ERROR;
	}
	rb_read = rb_count = 0;
  
	if ( (n_read = read(ttyfd, (char *)recv_buff, to_read)) <= 0)
	{
		if ( n_read < 0 && errno != EAGAIN)
		{
			rb_count = 0;
			//printf("GetMoreBytes: read error: %s\n", strerror(errno));
			return K2R_ERROR;
		}
		else
		{  
			/* Would have blocked, so we will select... */
			ptv = &tv;
			tv.tv_sec = toutms / 1000;
			tv.tv_usec = (toutms % 1000) * 1000;
      
			FD_ZERO( &r_set );
			FD_SET(ttyfd, &r_set);
      
      
			sel_ret = select( ttyfd+1, &r_set, NULL, NULL, ptv);
      
      
			if (sel_ret > 0 && FD_ISSET( ttyfd, &r_set ))
			{   
				/* Maybe there's something to receive */
				if ( (n_read = read(ttyfd, (char *)recv_buff, to_read)) < 0)
				{   /* We had a another booboo; wrap up and go home */
					printf("GetMoreBytes: read error: %s\n", strerror(errno));
					return K2R_ERROR;
				}
				rb_count = n_read;   /* Record how much is in recv_buff */
				if (gcfg_debug > 3)
					printf( "GMB: %d\n", rb_count);
				return K2R_NO_ERROR;
			}
			else if (sel_ret == 0)
			{   
				/* Select timed out */
			
				if (gcfg_debug > 3)
					printf( "GMB timeout\n");
				return K2R_TIMEOUT;
			}
			else
			{   
				/* Select had a booboo */
				printf("GetMoreBytes: select error: %s\n", strerror(errno));
				return K2R_ERROR;
			}
		}
    
	}
	/* We recv'ed ok the first time */
    rb_count = n_read;
    if (gcfg_debug > 3)
		
	{
      printf( "GMB: %d\n", rb_count);
	}
    return K2R_NO_ERROR;
  }



void edata2ew_SendStatus( unsigned char type, short ierr, char *note )
{
   MSG_LOGO    logo;
   char        msg[256];
   long        size;
   time_t        t;
 
/* Build the message
 *******************/ 
   logo.instid = InstId;
   logo.mod    = MyModId;
   logo.type   = type;

   time( &t );

   if( type == TypeHeartBeat )
   {
        sprintf( msg, "%ld %ld\n\0", (long) t, (long) MyPid);
   }
   else if( type == TypeError )
   {
        sprintf( msg, "%ld %hd %s\n\0", (long) t, ierr, note);
        logit( "et", "template: %s\n", note );
   }

   size = strlen( msg );   /* don't include the null byte in the message */     

/* Write the message to shared memory
 ************************************/
   if( tport_putmsg( &Region, &logo, size, msg ) != PUT_OK )
   {
        if( type == TypeHeartBeat ) {
           logit("et","template:  Error sending heartbeat.\n" );
        }
        else if( type == TypeError ) {
           logit("et","template:  Error sending error:%d.\n", ierr );
        }
   }

   return;
}





void printHexData (unsigned char *ptr, unsigned short n)
{
	unsigned short r;
	unsigned char c;
	unsigned char u;
	unsigned char *r_ptr;




	r_ptr = ptr;    // faccio la copia
    printf("    ");

	for (c=0; c<16; c++)
	{
		printf(" %02X", c);
	}
	printf("\r\n");


	for (c=0; c<(n/16); c++)
    {

		// offset 0000 (0)
		// offset 0010 (16)
		// offset 0020 (32)...

		// stampo l'offset
		printf("%04X ", c*16);

		// stampo i primi 16 bytes
		for (u=0; u<16; u++)
		{
				printf("%02X ", *r_ptr++);
		}

		printf(" ");
		r_ptr = ptr;    // faccio la copia
		for (u=0; u<16; u++)
		{
				if ((*ptr < 33) || (*ptr >126 ))
				{
						printf(".");
				}
				else
				{
						printf("%c", *ptr);
				}

				ptr++;
		}
		printf("\r\n");
		r_ptr += 16;
    }

	r_ptr = ptr;    // faccio la copia


   if (n%16>0)
	{
		printf("%04X ", c*16);
		for (u=0; u<(n%16); u++)
		{
				printf("%02X ", *r_ptr++);
		}

		for (u=0; u<(16-(n%16)); u++)
				printf("   ");
		printf(" ");


		r_ptr = ptr;    // faccio la copia
		for (u=0; u<(n%16); u++)
		{
				if ((*ptr < 33) || (*ptr >126 ))
				{
						printf(".");
				}
				else
				{
						printf("%c", *ptr);
				}

				ptr++;
		}
		printf("\r\n");

	}


	
}


void GPRMC_emulation ( char *ptr)
{
	unsigned char check, n;
	char *ptr2;
	char time[11];	// UTC time HHMMSS.mmm + null
	char flag;		// RMC flag
	char date[6];	// date DDMMYY + null
	char dummystring[10];
	unsigned char counter, idx;	
	char s = 0;
	struct tm timestamp;
	unsigned short dummy;	
	
	
	ptr2 = ptr;	// make a copy
	check = 0;
	n = 0;
	ptr2++;	
	while ((*ptr2 != '*') && (n<128))
	{
		check ^= *ptr2;
		//printf("%02X --> %c\r\n", check, *ptr2);
		ptr2++;
		n++;
	}
	//printf("checksum = %02X\r\n", check);
	ptr2++;
	sprintf(ptr2, "%02X\r", check);
	
	
	
}


char GPRMC_parsing ( char *ptr)
{
	unsigned char check, n;
	char *ptr2;
	char time[11];	// UTC time HHMMSS.mmm + null
	char flag;		// RMC flag
	char date[6];	// date DDMMYY + null
	char dummystring[10];
	unsigned char counter, idx;	
	char s = 0;
	struct tm timestamp;
	unsigned short dummy;
	
	ptr2 = ptr;	// make a copy
	
	
	
	
	if (strncmp("$GPRMC,",ptr, strlen("$GPRMC,")) != 0)
	{
		//printf("non sono uguali\r\n");
		return (-1);
	}
	
	check = 0;
	n = 0;
	ptr2++;
	while ((*ptr2 != '*') && (n<128))
	{
		
		check ^= *ptr2;
		//printf("%02X --> %c\r\n", check, *ptr2);
		ptr2++;
		n++;
		
	}
	//printf("%02X\r\n", check);
	sprintf(dummystring, "%02X", check);
	
	
	
	
	*ptr2++;	// asterisco
	
	if (strncmp(dummystring,ptr2, 2) != 0)
	{
//		printf("checksum fail...\r\n");
		return (-2);
	}
	else
	{
//		printf("NMEA checksum pass!\r\n");
	}
	
	
	// supponiamo abbia passato il checksum
	// adesso devo andare a estrarre i vari campi
	counter = 0;
	ptr += 7;	// seek to time field
	
	
	idx = 0;
	time[0] = 0;
	date[0] = 0;
	counter = 0;
	s = 0;
	while (*ptr != '*')
	{
	
		if (*ptr != ',')
		{
			switch (counter)
			{
				case 0:
						time[idx++] = *ptr;
						time[idx] = 0;
					break;
				case 8:
						date[idx++] = *ptr;
						date[idx] = 0;
					break;					
			}
		}
		else
		{
			counter++;
			idx = 0;
		}
		ptr++;
		
	}
	/*
	ptr = time;
	sprintf(ptr,"124511.154");
	sprintf(ptr,"124511.954");
	
	printf("got time %s UTC and date %s\r\n", time, date);*/
	
	
	// data
	ptr = date;
	strncpy(dummystring, ptr, 2);
	dummystring[2] = 0;
	timestamp.tm_mday = atoi(dummystring);
	
	ptr += 2;
	strncpy(dummystring, ptr, 2);
	dummystring[2] = 0;
	timestamp.tm_mon = atoi(dummystring);
	timestamp.tm_mon--;
	
	
	ptr += 2;
	strncpy(dummystring, ptr, 2);
	dummystring[2] = 0;
	timestamp.tm_year = atoi(dummystring);
	timestamp.tm_year += 2000;	
	timestamp.tm_year -= 1900;	
	
	
	//printf("%02d/%02d/%04d\r\n", timestamp.tm_mday, timestamp.tm_mon+1, timestamp.tm_year+1900);
	
	

	//printf("estrazione ora da %s\r\n", time);
	// hours
	ptr = time;
	strncpy(dummystring, ptr, 2);
	dummystring[2] = 0;
	timestamp.tm_hour = atoi(dummystring);
	
	// minutes
	ptr += 2;
	strncpy(dummystring, ptr, 2);
	dummystring[2] = 0;
	timestamp.tm_min = atoi(dummystring);
	
	// seconds
	ptr += 2;
	strncpy(dummystring, ptr, 2);
	dummystring[2] = 0;
	timestamp.tm_sec = atoi(dummystring);
	
	
	//printf("%02d:%02d:%02d\r\n", timestamp.tm_hour, timestamp.tm_min, timestamp.tm_sec);
	
	timestamp.tm_isdst = 0;
	
	
	
	/*printf("tm_sec = %02d\r\n", timestamp.tm_sec);
	printf("tm_min = %02d\r\n", timestamp.tm_min);
	printf("tm_hour = %02d\r\n", timestamp.tm_hour);
	printf("tm_mday = %02d\r\n", timestamp.tm_mday);
	printf("tm_mon = %02d\r\n", timestamp.tm_mon+1);
	printf("tm_year = %04d\r\n", timestamp.tm_year+1900);*/

	
	
	
	epoch = mktime(&timestamp);
	//printf("%lu\r\n", epoch);
	
	
	
	
	
	
	
	// let's check if milliseconds are present
	// sometimes GPS gives you time in format 182214.999 or 182214.99 for 182215.000 timestamp
	ptr += 3;
	
	switch (strlen(time))
	{
		case 10:
				// format 12.999 or 13.000				
				strncpy(dummystring, ptr, 3);
				dummy = atoi(dummystring);
				if (dummy > 500)
				{
						//printf("need to increase second\r\n");
						epoch++;
				}
					
			break;
		case 9:
				// format 12.99 or 13.00
				strncpy(dummystring, ptr, 2);
				dummy = atoi(dummystring);
				if (dummy > 50)
				{
						//printf("need to increase second\r\n");
						epoch++;
				}
				
			break;
	}
	return 0;
}
